@@ -1,10 +1,27 @@
+// Copyright 2021 MaidSafe.net limited.
+//
+// This SAFE Network Software is licensed to you under the MIT license <LICENSE-MIT
+// http://opensource.org/licenses/MIT> or the Modified BSD license <LICENSE-BSD
+// https://opensource.org/licenses/BSD-3-Clause>, at your option. This file may not be copied,
+// modified, or distributed except according to those terms. Please review the Licences for the
+// specific language governing permissions and limitations relating to use of the SAFE Network
+// Software.
+
 use super::{new_qp2p, new_qp2p_with_hcc, random_msg};
 use crate::utils;
 use anyhow::{anyhow, Result};
+use bytes::Bytes;
 use futures::future;
-use std::{collections::HashSet, time::Duration};
+use std::{
+    collections::{BTreeSet, HashSet},
+    time::Duration,
+};
+use tiny_keccak::{Hasher, Sha3};
 use tokio::time::timeout;
 
+/// SHA3-256 hash digest.
+type Digest256 = [u8; 32];
+
 #[tokio::test]
 async fn successful_connection() -> Result<()> {
     utils::init_logging();
@@ -40,7 +57,7 @@ async fn single_message() -> Result<()> {
 
     // Peer 2 connects and sends a message
     peer2.connect_to(&peer1_addr).await?;
-    let msg_from_peer2 = random_msg();
+    let msg_from_peer2 = random_msg(1024);
     peer2
         .send_message(msg_from_peer2.clone(), &peer1_addr)
         .await?;
@@ -76,7 +93,7 @@ async fn reuse_outgoing_connection() -> Result<()> {
 
     // Connect for the first time and send a message.
     alice.connect_to(&bob_addr).await?;
-    let msg0 = random_msg();
+    let msg0 = random_msg(1024);
     alice.send_message(msg0.clone(), &bob_addr).await?;
 
     // Bob should recieve an incoming connection and message
@@ -95,7 +112,7 @@ async fn reuse_outgoing_connection() -> Result<()> {
 
     // Try connecting again and send a message
     alice.connect_to(&bob_addr).await?;
-    let msg1 = random_msg();
+    let msg1 = random_msg(1024);
     alice.send_message(msg1.clone(), &bob_addr).await?;
 
     // Bob *should not* get an incoming connection since there is already a connection established
@@ -129,7 +146,7 @@ async fn reuse_incoming_connection() -> Result<()> {
 
     // Connect for the first time and send a message.
     alice.connect_to(&bob_addr).await?;
-    let msg0 = random_msg();
+    let msg0 = random_msg(1024);
     alice.send_message(msg0.clone(), &bob_addr).await?;
 
     // Bob should recieve an incoming connection and message
@@ -148,7 +165,7 @@ async fn reuse_incoming_connection() -> Result<()> {
 
     // Bob tries to connect to alice and sends a message
     bob.connect_to(&alice_addr).await?;
-    let msg1 = random_msg();
+    let msg1 = random_msg(1024);
     bob.send_message(msg1.clone(), &alice_addr).await?;
 
     // Alice *will not* get an incoming connection since there is already a connection established
@@ -252,10 +269,10 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
         anyhow!("No incoming connectino from Alice");
     }
 
-    let msg0 = random_msg();
+    let msg0 = random_msg(1024);
     alice.send_message(msg0.clone(), &bob_addr).await?;
 
-    let msg1 = random_msg();
+    let msg1 = random_msg(1024);
     bob.send_message(msg1.clone(), &alice_addr).await?;
 
     if let Some((src, message)) = alice_incoming_messages.next().await {
@@ -292,7 +309,7 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
         anyhow!("Unexpected incoming connection from {}", connecting_peer);
     }
 
-    let msg2 = random_msg();
+    let msg2 = random_msg(1024);
     bob.send_message(msg2.clone(), &alice_addr).await?;
 
     if let Some((src, message)) = alice_incoming_messages.next().await {
@@ -333,10 +350,10 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
     }
 
     // Send two messages, one from each end
-    let msg0 = random_msg();
+    let msg0 = random_msg(1024);
     alice.send_message(msg0.clone(), &bob_addr).await?;
 
-    let msg1 = random_msg();
+    let msg1 = random_msg(1024);
     bob.send_message(msg1.clone(), &alice_addr).await?;
 
     // Both messages are received  at the other end
@@ -357,6 +374,120 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
     Ok(())
 }
 
+fn hash(bytes: &Bytes) -> Digest256 {
+    let mut hasher = Sha3::v256();
+    let mut hash = Digest256::default();
+    hasher.update(bytes);
+    hasher.finalize(&mut hash);
+    hash
+}
+
+#[tokio::test]
+async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
+    use futures::future;
+
+    utils::init_logging();
+
+    let num_senders: usize = 10;
+    let num_messages_each: usize = 100;
+    let num_messages_total: usize = 1000;
+
+    let qp2p = new_qp2p()?;
+    let (server_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
+    let server_addr = server_endpoint.socket_addr();
+
+    let test_msgs: Vec<_> = (0..num_messages_each).map(|_| random_msg(1024)).collect();
+    let sending_msgs = test_msgs.clone();
+
+    let mut tasks = Vec::new();
+
+    // Receiver
+    tasks.push(tokio::spawn({
+        async move {
+            let mut num_received = 0;
+            let mut sending_tasks = Vec::new();
+
+            while let Some((src, msg)) = recv_incoming_messages.next().await {
+                log::info!("received from {:?} with message size {}", src, msg.len());
+                assert_eq!(msg.len(), test_msgs[0].len());
+
+                let sending_endpoint = server_endpoint.clone();
+
+                sending_tasks.push(tokio::spawn({
+                    async move {
+                        // Hash the inputs for couple times to simulate certain workload.
+                        let hash_result = hash(&msg);
+                        for _ in 0..5 {
+                            let _ = hash(&msg);
+                        }
+                        // Send the hash result back.
+                        sending_endpoint.connect_to(&src).await?;
+                        sending_endpoint
+                            .send_message(hash_result.to_vec().into(), &src)
+                            .await?;
+
+                        Ok::<_, anyhow::Error>(())
+                    }
+                }));
+
+                num_received += 1;
+                if num_received >= num_messages_total {
+                    break;
+                }
+            }
+
+            let _ = future::try_join_all(sending_tasks).await?;
+
+            Ok(())
+        }
+    }));
+
+    // Sender
+    for id in 0..num_senders {
+        let messages = sending_msgs.clone();
+        tasks.push(tokio::spawn({
+            let qp2p = new_qp2p()?;
+            let (send_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
+
+            async move {
+                let mut hash_results = BTreeSet::new();
+                log::info!("connecting {}", id);
+                send_endpoint.connect_to(&server_addr).await?;
+                for (index, message) in messages.iter().enumerate().take(num_messages_each) {
+                    let _ = hash_results.insert(hash(&message));
+                    log::info!("sender #{} sending message #{}", id, index);
+                    send_endpoint
+                        .send_message(message.clone(), &server_addr)
+                        .await?;
+                }
+
+                log::info!(
+                    "sender #{} completed sending messages, starts listening",
+                    id
+                );
+
+                while let Some((src, msg)) = recv_incoming_messages.next().await {
+                    log::info!(
+                        "#{} received from server {:?} with message size {}",
+                        id,
+                        src,
+                        msg.len()
+                    );
+                    assert!(hash_results.remove(&msg[..]));
+                    if hash_results.is_empty() {
+                        break;
+                    }
+                }
+
+                Ok::<_, anyhow::Error>(())
+            }
+        }));
+    }
+
+    let _ = future::try_join_all(tasks).await?;
+    Ok(())
+}
+
 #[tokio::test]
 async fn many_messages() -> Result<()> {
     use futures::future;
